@@ -1,8 +1,8 @@
 /*
  * cryptsetup-reencrypt - crypt utility for offline re-encryption
  *
- * Copyright (C) 2012-2018, Red Hat, Inc. All rights reserved.
- * Copyright (C) 2012-2018, Milan Broz All rights reserved.
+ * Copyright (C) 2012-2020 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2012-2020 Milan Broz All rights reserved.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -25,10 +25,9 @@
 #include <arpa/inet.h>
 #include <uuid/uuid.h>
 
-#define PACKAGE_REENC "crypt_reencrypt"
+#define PACKAGE_REENC "cryptsetup-reencrypt"
 
 #define NO_UUID "cafecafe-cafe-cafe-cafe-cafecafeeeee"
-#define MAX_BCK_SECTORS 8192
 
 static const char *opt_cipher = NULL;
 static const char *opt_hash = NULL;
@@ -43,7 +42,6 @@ static const char *opt_pbkdf = NULL;
 static long opt_pbkdf_memory = DEFAULT_LUKS2_MEMORY_KB;
 static long opt_pbkdf_parallel = DEFAULT_LUKS2_PARALLEL_THREADS;
 static long opt_pbkdf_iterations = 0;
-static int opt_version_mode = 0;
 static int opt_random = 0;
 static int opt_urandom = 0;
 static int opt_bsize = 4;
@@ -78,6 +76,7 @@ struct reenc_ctx {
 	uint64_t device_size_org_real;
 	uint64_t device_offset;
 	uint64_t device_shift;
+	uint64_t data_offset;
 
 	unsigned int stained:1;
 	unsigned int in_progress:1;
@@ -100,7 +99,6 @@ struct reenc_ctx {
 	} p[MAX_SLOT];
 	int keyslot;
 
-	struct timeval start_time, end_time;
 	uint64_t resume_bytes;
 };
 
@@ -147,7 +145,7 @@ static const char *luksType(const char *type)
 		return CRYPT_LUKS1;
 
 	if (!type || !strcmp(type, "luks"))
-		return DEFAULT_LUKS_FORMAT;
+		return crypt_get_default_type();
 
 	return NULL;
 }
@@ -196,8 +194,15 @@ static int device_check(struct reenc_ctx *rc, const char *device, header_magic s
 	ssize_t s;
 	uint16_t version;
 	size_t buf_size = pagesize();
+	struct stat st;
 
-	devfd = open(device, O_RDWR | O_EXCL | O_DIRECT);
+	if (stat(device, &st)) {
+		log_err(_("Cannot open device %s."), device);
+		return -EINVAL;
+	}
+
+	/* coverity[toctou] */
+	devfd = open(device, O_RDWR | (S_ISBLK(st.st_mode) ? O_EXCL : 0));
 	if (devfd == -1) {
 		if (errno == EBUSY) {
 			log_err(_("Cannot exclusively open %s, device in use."),
@@ -252,7 +257,7 @@ static int device_check(struct reenc_ctx *rc, const char *device, header_magic s
 		if (lseek(devfd, 0, SEEK_SET) == -1)
 			goto out;
 		s = write(devfd, buf, buf_size);
-		if (s < 0 || s != (ssize_t)buf_size) {
+		if (s < 0 || s != (ssize_t)buf_size || fsync(devfd) < 0) {
 			log_err(_("Cannot write device %s."), device);
 			r = -EIO;
 		}
@@ -269,55 +274,19 @@ out:
 	return r;
 }
 
-static int create_empty_header(const char *new_file, const char *old_file,
-			       uint64_t data_sector)
+static int create_empty_header(const char *new_file)
 {
-	struct stat st;
-	ssize_t size = 0;
 	int fd, r = 0;
-	char *buf;
 
-	/* Never create header > 4MiB */
-	if (data_sector > MAX_BCK_SECTORS)
-		data_sector = MAX_BCK_SECTORS;
+	log_dbg("Creating empty file %s of size 4096.", new_file);
 
-	/* new header file of the same size as old backup */
-	if (old_file) {
-		if (stat(old_file, &st) == -1 ||
-		    (st.st_mode & S_IFMT) != S_IFREG ||
-		    (st.st_size > 16 * 1024 * 1024))
-			return -EINVAL;
-		size = st.st_size;
-	}
+	/* coverity[toctou] */
+	fd = open(new_file, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IWUSR);
+	if (fd == -1 || posix_fallocate(fd, 0, 4096))
+		r = -EINVAL;
+	if (fd >= 0)
+		close(fd);
 
-	/*
-	 * if requesting key size change, try to use offset
-	 * here can be enough space to fit new key.
-	 */
-	if (opt_key_size && data_sector)
-		size = data_sector * SECTOR_SIZE;
-
-	/* if reducing size, be sure we have enough space */
-	if (opt_reduce_size)
-		size += opt_reduce_size;
-
-	log_dbg("Creating empty file %s of size %lu.", new_file, (unsigned long)size);
-
-	if (!size || !(buf = malloc(size)))
-		return -ENOMEM;
-	memset(buf, 0, size);
-
-	fd = creat(new_file, S_IRUSR|S_IWUSR);
-	if(fd == -1) {
-		free(buf);
-		return -EINVAL;
-	}
-
-	if (write(fd, buf, size) < size)
-		r = -EIO;
-
-	close(fd);
-	free(buf);
 	return r;
 }
 
@@ -476,9 +445,8 @@ static int activate_luks_headers(struct reenc_ctx *rc)
 	} else
 		return -EINVAL;
 
-	if ((r = crypt_init(&cd, rc->header_file_org)) ||
-	    (r = crypt_load(cd, CRYPT_LUKS, NULL)) ||
-	    (r = crypt_set_data_device(cd, rc->device)))
+	if ((r = crypt_init_data_device(&cd, rc->header_file_org, rc->device)) ||
+	    (r = crypt_load(cd, CRYPT_LUKS, NULL)))
 		goto out;
 
 	log_verbose(_("Activating temporary device using old LUKS header."));
@@ -487,9 +455,8 @@ static int activate_luks_headers(struct reenc_ctx *rc)
 		CRYPT_ACTIVATE_READONLY|CRYPT_ACTIVATE_PRIVATE)) < 0)
 		goto out;
 
-	if ((r = crypt_init(&cd_new, rc->header_file_new)) ||
-	    (r = crypt_load(cd_new, CRYPT_LUKS, NULL)) ||
-	    (r = crypt_set_data_device(cd_new, rc->device)))
+	if ((r = crypt_init_data_device(&cd_new, rc->header_file_new, rc->device)) ||
+	    (r = crypt_load(cd_new, CRYPT_LUKS, NULL)))
 		goto out;
 
 	log_verbose(_("Activating temporary device using new LUKS header."));
@@ -508,27 +475,24 @@ out:
 
 static int set_pbkdf_params(struct crypt_device *cd, const char *dev_type)
 {
+	const struct crypt_pbkdf_type *pbkdf_default;
 	struct crypt_pbkdf_type pbkdf = {};
 
-	if (!strcmp(dev_type, CRYPT_LUKS1)) {
-		if (opt_pbkdf && strcmp(opt_pbkdf, CRYPT_KDF_PBKDF2))
-			return -EINVAL;
-		pbkdf.type = CRYPT_KDF_PBKDF2;
-		pbkdf.hash = opt_hash ?: DEFAULT_LUKS1_HASH;
-		pbkdf.time_ms = opt_iteration_time ?: DEFAULT_LUKS1_ITER_TIME;
-	} else if (!strcmp(dev_type, CRYPT_LUKS2)) {
-		pbkdf.type = opt_pbkdf ?: DEFAULT_LUKS2_PBKDF;
-		pbkdf.hash = opt_hash ?: DEFAULT_LUKS1_HASH;
-		pbkdf.time_ms = opt_iteration_time ?: DEFAULT_LUKS2_ITER_TIME;
-		if (strcmp(pbkdf.type, CRYPT_KDF_PBKDF2)) {
-			pbkdf.max_memory_kb = opt_pbkdf_memory;
-			pbkdf.parallel_threads = opt_pbkdf_parallel;
-		}
-	} else
+	pbkdf_default = crypt_get_pbkdf_default(dev_type);
+	if (!pbkdf_default)
 		return -EINVAL;
 
+	pbkdf.type = opt_pbkdf ?: pbkdf_default->type;
+	pbkdf.hash = opt_hash ?: pbkdf_default->hash;
+	pbkdf.time_ms = (uint32_t)opt_iteration_time ?: pbkdf_default->time_ms;
+	if (strcmp(pbkdf.type, CRYPT_KDF_PBKDF2)) {
+		pbkdf.max_memory_kb = (uint32_t)opt_pbkdf_memory ?: pbkdf_default->max_memory_kb;
+		pbkdf.parallel_threads = (uint32_t)opt_pbkdf_parallel ?: pbkdf_default->parallel_threads;
+	}
+
 	if (opt_pbkdf_iterations) {
 		pbkdf.iterations = opt_pbkdf_iterations;
+		pbkdf.time_ms = 0;
 		pbkdf.flags |= CRYPT_PBKDF_NO_BENCHMARK;
 	}
 
@@ -569,6 +533,8 @@ static int create_new_header(struct reenc_ctx *rc, struct crypt_device *cd_old,
 			     const char *uuid,
 			     const char *key, int key_size,
 			     const char *type,
+			     uint64_t metadata_size,
+			     uint64_t keyslots_size,
 			     void *params)
 {
 	struct crypt_device *cd_new = NULL;
@@ -584,7 +550,19 @@ static int create_new_header(struct reenc_ctx *rc, struct crypt_device *cd_old,
 
 	r = set_pbkdf_params(cd_new, type);
 	if (r) {
-		log_err(_("Failed to set PBKDF parameters."));
+		log_err(_("Failed to set pbkdf parameters."));
+		goto out;
+	}
+
+	r = crypt_set_data_offset(cd_new, rc->data_offset);
+	if (r) {
+		log_err(_("Failed to set data offset."));
+		goto out;
+	}
+
+	r = crypt_set_metadata_size(cd_new, metadata_size, keyslots_size);
+	if (r) {
+		log_err(_("Failed to set metadata size."));
 		goto out;
 	}
 
@@ -704,6 +682,7 @@ static int backup_luks_headers(struct reenc_ctx *rc)
 	char cipher [MAX_CIPHER_LEN], cipher_mode[MAX_CIPHER_LEN];
 	char *key = NULL;
 	size_t key_size;
+	uint64_t mdata_size = 0, keyslots_size = 0;
 	int r;
 
 	log_dbg("Creating LUKS header backup for device %s.", hdr_device(rc));
@@ -729,14 +708,12 @@ static int backup_luks_headers(struct reenc_ctx *rc)
 	if (rc->reencrypt_mode == DECRYPT)
 		goto out;
 
-	if ((r = create_empty_header(rc->header_file_new, rc->header_file_org,
-		crypt_get_data_offset(cd))))
+	rc->data_offset = crypt_get_data_offset(cd) + ROUND_SECTOR(opt_reduce_size);
+
+	if ((r = create_empty_header(rc->header_file_new)))
 		goto out;
 
 	params.hash = opt_hash ?: DEFAULT_LUKS1_HASH;
-	params.data_alignment = crypt_get_data_offset(cd);
-	params.data_alignment += ROUND_SECTOR(opt_reduce_size);
-	params2.data_alignment = params.data_alignment;
 	params2.data_device = params.data_device = rc->device;
 	params2.sector_size = crypt_get_sector_size(cd);
 
@@ -768,6 +745,9 @@ static int backup_luks_headers(struct reenc_ctx *rc)
 	if (r < 0)
 		goto out;
 
+	if (isLUKS2(crypt_get_type(cd)) && crypt_get_metadata_size(cd, &mdata_size, &keyslots_size))
+		goto out;
+
 	r = create_new_header(rc, cd,
 		opt_cipher ? cipher : crypt_get_cipher(cd),
 		opt_cipher ? cipher_mode : crypt_get_cipher_mode(cd),
@@ -775,6 +755,8 @@ static int backup_luks_headers(struct reenc_ctx *rc)
 		key,
 		key_size,
 		rc->type,
+		mdata_size,
+		keyslots_size,
 		isLUKS2(rc->type) ? (void*)&params2 : (void*)&params);
 
 	if (!r && isLUKS2(rc->type))
@@ -792,16 +774,7 @@ static int backup_fake_header(struct reenc_ctx *rc)
 {
 	struct crypt_device *cd_new = NULL;
 	struct crypt_params_luks1 params = {0};
-	const struct crypt_pbkdf_type luks2_pbkdf = {
-		.type = DEFAULT_LUKS2_PBKDF,
-		.hash = opt_hash ?: DEFAULT_LUKS1_HASH,
-		.time_ms = DEFAULT_LUKS2_ITER_TIME,
-		.max_memory_kb = DEFAULT_LUKS2_MEMORY_KB,
-		.parallel_threads = DEFAULT_LUKS2_PARALLEL_THREADS
-	};
-	struct crypt_params_luks2 params2 = {
-		.pbkdf = &luks2_pbkdf
-	};
+	struct crypt_params_luks2 params2 = {0};
 	char cipher [MAX_CIPHER_LEN], cipher_mode[MAX_CIPHER_LEN];
 	const char *header_file_fake;
 	int r;
@@ -822,7 +795,7 @@ static int backup_fake_header(struct reenc_ctx *rc)
 		}
 	}
 
-	r = create_empty_header(header_file_fake, NULL, MAX_BCK_SECTORS);
+	r = create_empty_header(header_file_fake);
 	if (r < 0)
 		return r;
 
@@ -830,6 +803,7 @@ static int backup_fake_header(struct reenc_ctx *rc)
 	params2.data_alignment = params.data_alignment = 0;
 	params2.data_device = params.data_device = rc->device;
 	params2.sector_size = crypt_get_sector_size(NULL);
+	params2.pbkdf = crypt_get_pbkdf_default(CRYPT_LUKS2);
 
 	r = crypt_init(&cd_new, header_file_fake);
 	if (r < 0)
@@ -848,10 +822,12 @@ static int backup_fake_header(struct reenc_ctx *rc)
 		goto out;
 
 	/* The real header is backup header created in backup_luks_headers() */
-	if (rc->reencrypt_mode == DECRYPT)
+	if (rc->reencrypt_mode == DECRYPT) {
+		r = 0;
 		goto out;
+	}
 
-	r = create_empty_header(rc->header_file_new, rc->header_file_org, 0);
+	r = create_empty_header(rc->header_file_new);
 	if (r < 0)
 		goto out;
 
@@ -862,6 +838,8 @@ static int backup_fake_header(struct reenc_ctx *rc)
 		NULL, NULL,
 		(opt_key_size ? opt_key_size : DEFAULT_LUKS1_KEYBITS) / 8,
 		rc->type,
+		0,
+		0,
 		isLUKS2(rc->type) ? (void*)&params2 : (void*)&params);
 out:
 	crypt_free(cd_new);
@@ -894,7 +872,7 @@ static int restore_luks_header(struct reenc_ctx *rc)
 
 	/*
 	 * For new encryption and new detached header in file just move it.
-	 * For existing file try to ensure we have prealocated space for restore.
+	 * For existing file try to ensure we have preallocated space for restore.
 	 */
 	if (opt_new && rc->device_header) {
 		r = stat(rc->device_header, &st);
@@ -966,6 +944,8 @@ static int copy_data_forward(struct reenc_ctx *rc, int fd_old, int fd_new,
 
 	rc->resume_bytes = *bytes = rc->device_offset;
 
+	tools_reencrypt_progress(rc->device_size, *bytes, NULL);
+
 	if (write_log(rc) < 0)
 		return -EIO;
 
@@ -999,8 +979,8 @@ static int copy_data_forward(struct reenc_ctx *rc, int fd_old, int fd_new,
 		}
 
 		*bytes += (uint64_t)s2;
-		tools_time_progress(rc->device_size, *bytes,
-				    &rc->start_time, &rc->end_time);
+
+		tools_reencrypt_progress(rc->device_size, *bytes, NULL);
 	}
 
 	return quit ? -EAGAIN : 0;
@@ -1023,6 +1003,8 @@ static int copy_data_backward(struct reenc_ctx *rc, int fd_old, int fd_new,
 		*bytes = rc->resume_bytes;
 	}
 
+	tools_reencrypt_progress(rc->device_size, *bytes, NULL);
+
 	if (write_log(rc) < 0)
 		return -EIO;
 
@@ -1068,8 +1050,8 @@ static int copy_data_backward(struct reenc_ctx *rc, int fd_old, int fd_new,
 		}
 
 		*bytes += (uint64_t)s2;
-		tools_time_progress(rc->device_size, *bytes,
-				    &rc->start_time, &rc->end_time);
+
+		tools_reencrypt_progress(rc->device_size, *bytes, NULL);
 	}
 
 	return quit ? -EAGAIN : 0;
@@ -1083,7 +1065,7 @@ static void zero_rest_of_device(int fd, size_t block_size, void *buf,
 	log_dbg("Zeroing rest of device.");
 
 	if (lseek64(fd, offset, SEEK_SET) < 0) {
-		log_dbg(_("Cannot seek to device offset.\n"));
+		log_dbg("Cannot seek to device offset.");
 		return;
 	}
 
@@ -1095,9 +1077,9 @@ static void zero_rest_of_device(int fd, size_t block_size, void *buf,
 			s1 = *bytes;
 
 		s2 = write(fd, buf, s1);
-		if (s2 < 0) {
-			log_dbg("Write error, expecting %zu, got %zd.",
-				block_size, s2);
+		if (s2 != s1) {
+			log_dbg("Write error, expecting %zd, got %zd.",
+				s1, s2);
 			return;
 		}
 
@@ -1156,8 +1138,6 @@ static int copy_data(struct reenc_ctx *rc)
 	}
 
 	set_int_handler(0);
-	tools_time_progress(rc->device_size, bytes,
-			    &rc->start_time, &rc->end_time);
 
 	if (rc->reencrypt_direction == FORWARD)
 		r = copy_data_forward(rc, fd_old, fd_new, block_size, buf, &bytes);
@@ -1174,9 +1154,7 @@ static int copy_data(struct reenc_ctx *rc)
 
 	set_int_block(1);
 
-	if (r == -EAGAIN)
-		 log_err(_("Interrupted by a signal."));
-	else if (r < 0)
+	if (r < 0 && r != -EAGAIN)
 		log_err(_("IO error during reencryption."));
 
 	(void)write_log(rc);
@@ -1302,18 +1280,23 @@ static int init_keyfile(struct reenc_ctx *rc, struct crypt_device *cd, int slot_
 	if (r < 0)
 		return r;
 
-	r = crypt_activate_by_passphrase(cd, NULL, slot_check, password,
-					 passwordLen, 0);
+	/* mode ENCRYPT call this without header */
+	if (cd) {
+		r = crypt_activate_by_passphrase(cd, NULL, slot_check, password,
+						 passwordLen, 0);
 
-	/*
-	 * Allow keyslot only if it is last slot or if user explicitly
-	 * specify which slot to use (IOW others will be disabled).
-	 */
-	if (r >= 0 && opt_key_slot == CRYPT_ANY_SLOT &&
-	    crypt_keyslot_status(cd, r) != CRYPT_SLOT_ACTIVE_LAST) {
-		log_err(_("Key file can be used only with --key-slot or with "
-			  "exactly one key slot active."));
-		r = -EINVAL;
+		/*
+		 * Allow keyslot only if it is last slot or if user explicitly
+		 * specify which slot to use (IOW others will be disabled).
+		 */
+		if (r >= 0 && opt_key_slot == CRYPT_ANY_SLOT &&
+		    crypt_keyslot_status(cd, r) != CRYPT_SLOT_ACTIVE_LAST) {
+			log_err(_("Key file can be used only with --key-slot or with "
+				  "exactly one key slot active."));
+			r = -EINVAL;
+		}
+	} else {
+		r = slot_check == CRYPT_ANY_SLOT ? 0 : slot_check;
 	}
 
 	if (r < 0) {
@@ -1340,20 +1323,22 @@ static int initialize_passphrase(struct reenc_ctx *rc, const char *device)
 	log_dbg("Passphrases initialization.");
 
 	if (rc->reencrypt_mode == ENCRYPT && !rc->in_progress) {
-		r = init_passphrase1(rc, cd, _("Enter new passphrase: "), opt_key_slot, 0, 1);
+		if (opt_key_file)
+			r = init_keyfile(rc, NULL, opt_key_slot);
+		else
+			r = init_passphrase1(rc, NULL, _("Enter new passphrase: "), opt_key_slot, 0, 1);
 		return r > 0 ? 0 : r;
 	}
 
-	if ((r = crypt_init(&cd, device)) ||
-	    (r = crypt_load(cd, CRYPT_LUKS, NULL)) ||
-	    (r = crypt_set_data_device(cd, rc->device))) {
+	if ((r = crypt_init_data_device(&cd, device, rc->device)) ||
+	    (r = crypt_load(cd, CRYPT_LUKS, NULL))) {
 		crypt_free(cd);
 		return r;
 	}
 
 	if (opt_key_slot != CRYPT_ANY_SLOT)
 		snprintf(msg, sizeof(msg),
-			 _("Enter passphrase for key slot %u: "), opt_key_slot);
+			 _("Enter passphrase for key slot %d: "), opt_key_slot);
 	else
 		snprintf(msg, sizeof(msg), _("Enter any existing passphrase: "));
 
@@ -1364,7 +1349,7 @@ static int initialize_passphrase(struct reenc_ctx *rc, const char *device)
 		   rc->reencrypt_mode == DECRYPT) {
 		r = init_passphrase1(rc, cd, msg, opt_key_slot, 1, 0);
 	} else for (i = 0; i < crypt_keyslot_max(crypt_get_type(cd)); i++) {
-		snprintf(msg, sizeof(msg), _("Enter passphrase for key slot %u: "), i);
+		snprintf(msg, sizeof(msg), _("Enter passphrase for key slot %d: "), i);
 		r = init_passphrase1(rc, cd, msg, i, 1, 0);
 		if (r == -ENOENT) {
 			r = 0;
@@ -1605,6 +1590,11 @@ static void help(poptContext popt_context,
 	if (key->shortName == '?') {
 		log_std("%s %s\n", PACKAGE_REENC, PACKAGE_VERSION);
 		poptPrintHelp(popt_context, stdout, 0);
+		poptFreeContext(popt_context);
+		exit(EXIT_SUCCESS);
+	} else if (key->shortName == 'V') {
+		log_std("%s %s\n", PACKAGE_REENC, PACKAGE_VERSION);
+		poptFreeContext(popt_context);
 		exit(EXIT_SUCCESS);
 	} else
 		usage(popt_context, EXIT_SUCCESS, NULL, NULL);
@@ -1616,11 +1606,11 @@ int main(int argc, const char **argv)
 		{ NULL,    '\0', POPT_ARG_CALLBACK, help, 0, NULL,                         NULL },
 		{ "help",  '?',  POPT_ARG_NONE,     NULL, 0, N_("Show this help message"), NULL },
 		{ "usage", '\0', POPT_ARG_NONE,     NULL, 0, N_("Display brief usage"),    NULL },
+		{ "version",'V', POPT_ARG_NONE,     NULL, 0, N_("Print package version"),  NULL },
 		POPT_TABLEEND
 	};
 	static struct poptOption popt_options[] = {
 		{ NULL,                '\0', POPT_ARG_INCLUDE_TABLE, popt_help_options, 0, N_("Help options:"), NULL },
-		{ "version",           '\0', POPT_ARG_NONE, &opt_version_mode,          0, N_("Print package version"), NULL },
 		{ "verbose",           'v',  POPT_ARG_NONE, &opt_verbose,               0, N_("Shows more detailed error messages"), NULL },
 		{ "debug",             '\0', POPT_ARG_NONE, &opt_debug,                 0, N_("Show debug messages"), NULL },
 		{ "block-size",        'B',  POPT_ARG_INT, &opt_bsize,                  0, N_("Reencryption block size"), N_("MiB") },
@@ -1673,12 +1663,6 @@ int main(int argc, const char **argv)
 		usage(popt_context, EXIT_FAILURE, poptStrerror(r),
 		      poptBadOption(popt_context, POPT_BADOPTION_NOALIAS));
 
-	if (opt_version_mode) {
-		log_std("%s %s\n", PACKAGE_REENC, PACKAGE_VERSION);
-		poptFreeContext(popt_context);
-		exit(EXIT_SUCCESS);
-	}
-
 	if (!opt_batch_mode)
 		log_verbose(_("Reencryption will change: %s%s%s%s%s%s."),
 			opt_keep_key ? "" :  _("volume key"),
@@ -1706,12 +1690,12 @@ int main(int argc, const char **argv)
 
 	if (opt_pbkdf && crypt_parse_pbkdf(opt_pbkdf, &opt_pbkdf))
 		usage(popt_context, EXIT_FAILURE,
-		_("Password-based key derivation function (PBKDF) can be only pbkdf2 or argon2i/argon2id.\n"),
+		_("Password-based key derivation function (PBKDF) can be only pbkdf2 or argon2i/argon2id."),
 		poptGetInvocationName(popt_context));
 
 	if (opt_pbkdf_iterations && opt_iteration_time)
 		usage(popt_context, EXIT_FAILURE,
-		_("PBKDF forced iterations cannot be combined with iteration time option.\n"),
+		_("PBKDF forced iterations cannot be combined with iteration time option."),
 		poptGetInvocationName(popt_context));
 
 	if (opt_bsize < 1 || opt_bsize > 64)
